"""
Authors: Christopher Re and Ben Recht
Copyright 2010-2011 Christopher Re and Ben Recht

This code is released under the GPLv3 License.
http://www.gnu.org/licenses/gpl.html

The files in this directory and its subdirectories are part of Jellyfish.

Jellyifsh is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Jellyfish is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Jellyfish.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy, collections

cdef extern from "examples.h":
    cdef struct example:
        pass
    # pass by refernce
    cdef example* load_examples(char*szName, int nRows, int nCols, int nExamples)


cdef class ExampleHolder:
    cdef example* examples
    cdef int nRows, nCols, nExamples
    def __cinit__(self, char* szFileName):
        cdef int nRows = 0, nCols = 0, nExamples = 0
        self.examples  = load_examples(szFileName, nRows, nCols, nExamples)
        self.nRows     = nRows
        self.nCols     = nCols
        self.nExamples = nExamples

    cdef example* get_examples(self):
        return self.examples

cdef extern from "fvector.h":
    cdef cppclass FVector:
        FVector()
        FVector(int)
        void zero()
        void set(int,double)
        float get(int)
        int get_size()
        void set_size(int)
    cdef FVector* allocate_fvector_array(int) 

cdef numpy_array_of_vector(FVector f):
    return numpy.array([f.get(i) for i in range(f.get_size())])




cdef extern from "model.h":
    cdef struct model:
        int nRows
        int nCols
        double moviemean
        FVector *L, *R
    cdef model* build_model(int, int, double, FVector*, FVector*)

cdef model* model_from_gmodel(gmodel):
    cdef FVector fv
    cdef int i,j
    rank = gmodel.L[0].size
    fv.set_size(rank)

    nRows = gmodel.nRows
    nCols = gmodel.nCols
  
    L = allocate_fvector_array( gmodel.nRows )
    R = allocate_fvector_array( gmodel.nCols )

    for i in range(nRows):
        for j in xrange(rank):
            L[i].set(j, gmodel.L[i][j])

    for i in xrange(nCols):
        for j in xrange(rank):
            R[i].set(j, gmodel.R[i][j])

    moviemean = gmodel.moviemean
    return build_model(nRows, nCols, moviemean, L, R)

# You can create a test handler from Python, but you cannot do anything with it.
cdef extern from "test_handler.h":
    cdef cppclass test_handler:
        test_handler(char *)
        double test(model)

cdef class TestHandler:
    cdef test_handler* t
    def __cinit__(self, char*szFile):
        self.t = new test_handler(szFile)
    cdef test_handler* get_handler(self):
        return self.t

    cpdef double test(self,m):
        return self.t.test(model_from_gmodel(m)[0])
    

cdef extern from "cycle_consensus_threads_with_example.h":
    cdef struct training_parameters:
        example* examples
        int nRows, nCols, nExamples
        test_handler *t
        int nEpochs, max_rank
        int bTraceNorm
        int nSplits, nShufflers
        int seed
        int bPrint
        float initial_step_size, step_diminish
        float B, mu
        model* initial_guess
    cdef model superfast_jellyfish(training_parameters)

tps    = collections.namedtuple("tps", "example_holder test_handler nEpochs max_rank bTraceNorm nSplits nShufflers seed initial_stepsize step_diminish b mu")
gmodel = collections.namedtuple("gmodel", "nRows nCols moviemean L R")
cdef numpify_model(model m):
    return gmodel(nRows = m.nRows, nCols = m.nCols, moviemean = m.moviemean, 
                  L=numpy.array([numpy_array_of_vector(m.L[i]) for i in range(m.nRows)]),
                  R=numpy.array([numpy_array_of_vector(m.R[i]) for i in range(m.nCols)]))

cdef _internal_jellyfish(tps, initial_guess):
    cdef training_parameters to_f
    cdef ExampleHolder eh = tps.example_holder
    cdef TestHandler th

    to_f.examples    = eh.get_examples()
    to_f.nRows       = eh.nRows
    to_f.nCols       = eh.nCols
    to_f.nExamples   = eh.nExamples

    if tps.test_handler is None:
        to_f.t  = NULL
    else:
        th             = tps.test_handler
        to_f.t         = th.get_handler()

    to_f.nEpochs       = tps.nEpochs
    to_f.max_rank      = tps.max_rank
    to_f.bTraceNorm    = tps.bTraceNorm
    to_f.nSplits       = tps.nSplits
    to_f.nShufflers    = tps.nShufflers
    to_f.seed          = tps.seed
    to_f.initial_step_size = tps.initial_stepsize
    to_f.step_diminish = tps.step_diminish
    to_f.B             = tps.b
    to_f.mu            = tps.mu
    if initial_guess is None:
        to_f.initial_guess = NULL
    else:
        to_f.initial_guess = model_from_gmodel(initial_guess)
    return numpify_model(superfast_jellyfish(to_f))

def call_jellyfish(tps, initial_guess=None):
    return _internal_jellyfish(tps,initial_guess)
