/* Authors: Christopher Re and Ben Recht
Copyright 2010-2011 Christopher Re and Ben Recht

This code is released under the GPLv3 License.
http://www.gnu.org/licenses/gpl.html

The files in this directory and its subdirectories are part of Jellyfish.

Jellyifsh is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Jellyfish is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Jellyfish.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <assert.h>
#include <cstdlib>
#include <cstring>
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include "global_macros.h"
#include "timer.h"
#include "examples.h"
#include "fvector.h"
#include "pthread.h"
#include "simple_random.h"
#include "cycle_consensus_threads_with_example.h"

#include <vector>
using namespace std;
class FVector;

typedef vector<struct example> example_vec;
typedef vector<struct example>::iterator example_vec_it;



void
inline
find_bin(int row,        int col, 
	 int row_divide, int col_divide, 
	 int &k,         int &l) {
  k = (row >= row_divide) ? 1 : 0;
  l = (col >= col_divide) ? 1 : 0;
}

void 
permute(simple_random &rand, int *d, int n) {
  for(int i = n-1; i > 0; i--) {
    int rand_index = rand.rand_int(i); // pick something [0,i]
    int temp       = d[i];
    d[i]           = d[rand_index];
    d[rand_index]  = temp;    
  }
  VERBOSE_ONLY(
	       for(int i =0; i < n; i++) {
		 cout << i << " -> " << d[i] << endl;
	       }
	       )
}

// builds and returns a ragged array that contains all the vectors sliced up.
// slice_size is the size of all but the last slice (0 if there is one slice). 
// The last_slice may be irregular since nRows % nSlices != 0.
struct vector_slice {
  FVector **vecs;
  int nVecs, nSlices, slice_size, last_slice_size; 
  // This is an optional (possibly NULL) double pointer that is a ragged array storing  
  int **degree_info;
  vector_slice() {degree_info = NULL; vecs=NULL; nSlices = 0;}
  ~vector_slice() {
    for(int i = 0; i < nSlices; i++) {
      delete [] vecs[i];
      if(degree_info != NULL) {delete degree_info[i];}
    }
    if(vecs != NULL)        { delete [] vecs;        }
    if(degree_info != NULL) { delete [] degree_info; }
  }
};
ostream& operator<<(ostream &out, const vector_slice &vs) {
  // TODO: Add whether degree info is not null
  out << "[nVecs=" << vs.nVecs << " nSlices=" << vs.nSlices << " slice_size = " << vs.slice_size << " last_slice_size=" << vs.last_slice_size << "]";
  return out;
}

void
build_slices(int nVecs, int nSlices, vector_slice &vs, bool _compute_degree) {
  assert(nSlices > 0);
  vs.nVecs      = nVecs;
  vs.vecs       = new FVector*[nSlices];
  if(_compute_degree) {vs.degree_info = new int*[nSlices];}

  vs.nSlices    = nSlices;
  vs.slice_size = nVecs/nSlices;
  // allocate the individual slices (all but the last one)
  for(int i = 0; i < nSlices - 1; i++) {
    vs.vecs[i] = new FVector[vs.slice_size];    
    if(_compute_degree) { vs.degree_info[i] = new int[vs.slice_size]; }
  }
  // allocate the last slice
  vs.last_slice_size = nVecs - (nSlices - 1)* vs.slice_size;
  vs.vecs[nSlices - 1] = new FVector[vs.last_slice_size];
  if(_compute_degree) { vs.degree_info[nSlices - 1] = new int[vs.last_slice_size]; }
}

void find_slice(const vector_slice &vs, int v, int &slice_index, int &slice_offset) {
  if(v > vs.nVecs - vs.last_slice_size) { 
    slice_index  = vs.nSlices - 1;
    slice_offset = v - vs.slice_size * (vs.nSlices - 1);
  } else {
    slice_index = v/vs.slice_size;
    slice_offset = v - vs.slice_size * slice_index;
  }
  VERBOSE_ONLY(cout << vs << " " << v << " to " << slice_index << "," << slice_offset << endl;);
}

void store_vec(vector_slice &vs, int slice_index, int slice_offset, FVector &f) {
  vs.vecs[slice_index][slice_offset] = f;
}

void store_vec_degree(vector_slice &vs, int slice_index, int slice_offset, FVector &f, int degree) {
    vs.vecs[slice_index][slice_offset] = f;
    vs.degree_info[slice_index][slice_offset] = degree;
}
 
FVector // should this be FVector&
get_vec(const vector_slice &vs, int slice_index, int slice_offset) {
  return vs.vecs[slice_index][slice_offset];
}

// int
// get_degree(const vector_slice &vs, int slice_index, int slice_offset) {
//   return vs.degree_info[slice_index][slice_offset];
// }


// *****************************
// *THREADING
struct gradient_thread_info {
  int id;
  FVector *cur_L, *cur_R;
  
  example_vec &ev; 
  double moviemean, cur_step_size;
  // Optimization problem
  bool tracenorm; // false we do ball project
  // *************************************
  // Information for proximal points.
  int *L_degree, *R_degree;
  double B, mu;
  gradient_thread_info(int _id, FVector *L, int *_L_degree, FVector *R, int *_R_degree, 
		       example_vec &v, double _moviemean, double _cur_step_size, double _B, double _mu, bool _tracenorm) : ev(v) 
  { id = _id; cur_L = L; L_degree = _L_degree, cur_R = R; R_degree = _R_degree; 
    moviemean = _moviemean; cur_step_size = _cur_step_size; B = _B; mu = _mu;
    tracenorm = _tracenorm;
  }
};
void*
gradient_thread_main(void * params) {
  gradient_thread_info *gti = (struct gradient_thread_info*) params;
  // read the params
  DEBUG_ONLY(int id               = gti->id;)
  FVector *cur_L       = gti->cur_L;
  FVector *cur_R       = gti->cur_R;
  example_vec &ev      = gti->ev;
  double moviemean     = gti->moviemean;
  double cur_step_size = gti->cur_step_size;
  double B             = gti->B;
  double mu            = gti->mu;
  bool  _tracenorm     = gti->tracenorm;

  int     *L_degree    = gti->L_degree;
  int     *R_degree    = gti->R_degree;  

  // Make sure that we are good.
  assert((R_degree != NULL && L_degree != NULL) || !_tracenorm);

  // get to work.
  DEBUG_ONLY(timer part_timer(true););
  double B_squared     = B*B;
  FVector _tempL;
  
  if(_tracenorm) {
     for(example_vec_it it = ev.begin();
	 it != ev.end(); it++) {
       int row_index = it->row;
       int col_index = it->col;
       double rating = it->rating;
       
       //    err = sum(L(row_idx,:).*R(col_idx,:),2)+moviemean-double(rating);
       double err     = FVector::dot(cur_L[row_index],cur_R[col_index]) + moviemean - rating;
    
       // E = double(stepsize(epoch)*err);	
       double e       = -(cur_step_size * err);


     _tempL = cur_L[row_index];
     _tempL.scale(1 - mu*cur_step_size/((double) L_degree[row_index]));
     _tempL.scale_and_add(cur_R[col_index],e);	 
     
     cur_R[col_index].scale(1 - mu*cur_step_size/((double) R_degree[col_index])); 
     cur_R[col_index].scale_and_add(cur_L[row_index],e);	      
     // Store back L
     cur_L[row_index] = _tempL;
     }

  } else {     
    for(example_vec_it it = ev.begin();
	it != ev.end(); it++) {
      int row_index = it->row;
      int col_index = it->col;
      double rating = it->rating;
	
      //    err = sum(L(row_idx,:).*R(col_idx,:),2)+moviemean-double(rating);
      double err     = FVector::dot(cur_L[row_index],cur_R[col_index]) + moviemean - rating;
      
	// E = double(stepsize(epoch)*err);	
      double e       = -(cur_step_size * err);
	
      _tempL = cur_L[row_index];
      _tempL.scale_and_add(cur_R[col_index],e);	 
      cur_R[col_index].scale_and_add(cur_L[row_index],e);      
      // Store back L
	cur_L[row_index] = _tempL;
	
	// Do the projection if needed
	cur_L[row_index].ball_project(B, B_squared);	  
	cur_R[col_index].ball_project(B, B_squared);
      }
  }
  DEBUG_ONLY(cout << " part id=" << id << " took " << part_timer << " on " << ev.size() << " examples " << endl;);
  delete gti;
  return NULL;
}

struct example_thread_info {
  int nRow_Slices, nCol_Slices;
  int nRows, nCols, nExamples;
  int *row_perm, *col_perm, *example_perm;
  struct example* examples;
  vector_slice &L_slice, &R_slice;
  example_vec **example_vectors;
  simple_random &rand;
  example_thread_info(int nRow_Slices, int nCol_Slices, 
		      int nRows, int nCols, int nExamples,
		      int *row_perm, int *col_perm, int* example_perm,		       		      
		      struct example* examples,
		      vector_slice &L_slice, vector_slice &R_slice,
		      example_vec **example_vectors, simple_random &_rand)  :
  nRow_Slices(nRow_Slices), nCol_Slices(nCol_Slices), 
  nRows(nRows), nCols(nCols), nExamples(nExamples), 
  row_perm(row_perm), col_perm(col_perm), example_perm(example_perm),
  examples(examples),
  L_slice(L_slice), R_slice(R_slice), example_vectors(example_vectors), rand(_rand)
  { }
};


void*
example_handler_thread_main(void *params) {
  // Unpack the parameters
  example_thread_info *eti = (example_thread_info*) params;
  int nRow_Slices          = eti->nRow_Slices;
  int nCol_Slices          = eti->nCol_Slices;
  int nRows                = eti->nRows;
  int nCols                = eti->nCols;
  int nExamples            = eti->nExamples;
  int *row_perm            = eti->row_perm;
  int *col_perm            = eti->col_perm;
  int *example_perm        = eti->example_perm;
  simple_random &_rand     = eti->rand;
  struct example* examples = eti->examples;

  vector_slice &L_slice = eti->L_slice;
  vector_slice &R_slice = eti->R_slice;
  example_vec**example_vectors = eti->example_vectors;
  // Create a random permutation
  DEBUG_ONLY(timer perm_timer(true););
  
  
  // ********** Construct the permutations ******
  permute(_rand, row_perm, nRows);
  permute(_rand, col_perm, nCols);
  permute(_rand, example_perm, nExamples);
  
  DEBUG_ONLY(cout << "\t Computing Permutation took " << perm_timer << endl; );
 
  // ********** MOVE THE EXAMPLES *******
  DEBUG_ONLY(timer e_timer(true););
  // send the examples to their slice homes
  for(int k = 0; k < nRow_Slices; k++) {
    for(int l = 0; l < nCol_Slices; l++) {
      example_vectors[k][l].clear();
    }
  }
  DEBUG_ONLY(cout << "\t Example Holders Cleared " << endl; );
  // send the examples to their slice homes
  for(int i = 0; i < nExamples; i++) {
    // permute the rows and columns
    int perm_i = example_perm[i];
    struct example &example_ref = examples[perm_i];

    int r = row_perm[example_ref.row];
    int c = col_perm[example_ref.col];
    // now build a new example with row and column offsets set appropriatly.
    struct example e;
    int k = 0, l = 0;
    // NB: find slice is **ONLY** using the block partitionining information
    // the vectors are invalid here.
    find_slice(L_slice,r, k, e.row);
    find_slice(R_slice,c, l, e.col);
    e.rating = example_ref.rating;
    example_vectors[k][l].push_back(e);
  }

  DEBUG_ONLY(cout << "\t Examples Shuffled in " << e_timer << endl;);
  // NB: We should not free the etis because they can be reused.
  return NULL;
}

int advance_eti(int current_eti_index, int number_of_etis) {
  return (current_eti_index + 1) % number_of_etis;
}

int*
init_permutation(int nSize) {
  int *ret = new int[nSize];
  for(register int i = nSize - 1; i >= 0; i--) { ret[i] = i; }
  return ret;
}

// I'm not sure why it's called this either, but I hate that iTunes ate this album.
struct model
//szTrainFile, t, epochs, max_rank, initial_stepsize, stepsize_diminish, nRowSlices, nColSlices
superfast_jellyfish(struct training_parameters tp) {
  test_handler *t = tp.t;
  int nEpochs = tp.nEpochs;
  int max_rank =tp.max_rank;
  double initial_step_size = tp.initial_step_size, step_diminish = tp.step_diminish;  
  double divergent_series_rule = 0.0;
  bool bTraceNorm = tp.bTraceNorm;
  double B = tp.B;
  double mu = tp.mu;
  int nShufflers = tp.nShufflers;
  int nRow_Slices = tp.nSplits;
  int nCol_Slices = tp.nSplits; 
  


  assert(nRow_Slices > 0 && nCol_Slices > 0);
  // parameters
  double cur_step_size = initial_step_size;
  cout << "# [superfast]  Parameters:" << endl;
  cout << "# [superfast]  nepochs=" << nEpochs << " B=" << B << " max_rank=" << max_rank << " nSplits=" << tp.nSplits << " nShufflers=" << nShufflers << endl;
  cout << "# [superfast]  initial_stepsize=" << initial_step_size << " step_diminish=" << step_diminish << " divergent_rule=" << divergent_series_rule << endl;
  cout << "# [superfast]  tracenorm=" << bTraceNorm << " mu=" << mu << endl;
  int nSteps = 0;
  timer whole_jellyfish(true);
  
  DEBUG_ONLY(timer data_load; data_load.start(););
  // set by load examples
  int nRows = tp.nRows, nCols = tp.nCols, nExamples = tp.nExamples; 
  struct example *examples    = tp.examples;
  cout << "# [superfast]  nExamples=" << nExamples << " nRows=" << nRows << " nCols=" << nCols << endl;
  // Create the Test Handler
  assert(examples != NULL);
  //cout << " \t tp.examples=" << tp.examples << endl;

  test_handler  train_tester(examples, nExamples);

  
  // ***
  // intialize L and R
  // **************************
  // Divide up the data
  // Need a double array of FVectors(30)s L, R
  FVector::_default_n = max_rank; // this helps make allocation easy. 
 
  // TODO: Remove these for an extra copy.
  // Build the model
  struct model m, *initial_guess;
  initial_guess = tp.initial_guess;
  FVector *L,*R;
  if(initial_guess == NULL) {
    m.nRows = nRows; m.nCols = nCols;
    L  = new FVector[nRows];
    R  = new FVector[nCols];
  } else {
    m.nRows = nRows; m.nCols = nCols;
    assert(initial_guess->nRows == nRows);
    assert(initial_guess->nCols == nCols);
    L = initial_guess->L;
    R = initial_guess->R;
    m.moviemean = initial_guess->moviemean;
  }
  m.L = L;
  m.R = R;

  // *** Compute Stats about the Examples
  double moviemean = 0.0;
  int *L_degree = NULL, *R_degree = NULL;
  if(bTraceNorm) { 
    // the outdegree of each row
    L_degree = new int[nRows]; bzero(L_degree, sizeof(int)*nRows); 
    R_degree = new int[nCols]; bzero(R_degree, sizeof(int)*nCols); 
  }
  for(int i = 0; i < nExamples; i++) {
    if(bTraceNorm) { // TODO: Hoist this computation.
      L_degree[examples[i].row] ++;
      R_degree[examples[i].col] ++;
    } else {
      moviemean += examples[i].rating;
    }
  }

  moviemean /= ((double) nExamples);
  if(initial_guess == NULL) {m.moviemean = moviemean;}
 
  DEBUG_ONLY(data_load.stop(););
  DEBUG_ONLY(cout  << "Data Loaded: N=" << nExamples << " c=" << nCols << " r=" << nRows << " in " << data_load.elapsed() << "s" << endl;);

  
  // Build slice structures
  vector_slice L_slice, R_slice;
  build_slices(nRows, nRow_Slices, L_slice, bTraceNorm);
  build_slices(nCols, nCol_Slices, R_slice, bTraceNorm);
  DEBUG_ONLY(cout  << "Slices built" << endl;);

  // **** 
  // permutation in each row and col

  int number_of_etis = nShufflers+1;
  int *row_perms[number_of_etis];
  int *col_perms[number_of_etis];
  int *example_perms[number_of_etis];
  example_vec** all_example_vectors[number_of_etis]; 
  example_thread_info* etis[number_of_etis];

  for(int i = 0; i < number_of_etis; i++) {
    row_perms[i] = init_permutation(nRows);
    col_perms[i] = init_permutation(nCols);
    example_perms[i] = init_permutation(nExamples);
    all_example_vectors[i] = new example_vec*[nRow_Slices];
    for(int j = 0; j < nRow_Slices; j++) {
      all_example_vectors[i][j] = new example_vec[nCol_Slices];
    }

    unsigned short *c_seed = NULL;
    if(tp.seed) {
      c_seed = new unsigned short[3];
      c_seed[0] = c_seed[1] = c_seed[2] = tp.seed;
    }

    simple_random * rand_i = new simple_random(c_seed);

    etis[i] = new example_thread_info(nRow_Slices, nCol_Slices, 
				      nRows, nCols, nExamples,
				      row_perms[i], col_perms[i], example_perms[i], 
				      examples,
				      L_slice, R_slice, 
				      all_example_vectors[i], *rand_i);
  }
  DEBUG_ONLY(cout  << "Example Structures Built" << endl;);
 
  // build the example data structures  
  
  int current_eti_index = 0;
  
  pthread_t example_threads[number_of_etis];
  bool      thread_running[number_of_etis];
  for(int j = 0; j < number_of_etis; j++) {
    pthread_create( &example_threads[j], NULL, 
		    example_handler_thread_main, (void*) etis[j] );
  }
  
  DEBUG_ONLY(cout  << "Model built. Ready to start." << endl;);
  for(int epoch = 0; epoch < nEpochs; epoch++) {
    DEBUG_ONLY(cout << "Current eti: " << current_eti_index << endl;);
    // Get the data (join on the data thread)
    timer wait_timer(true);
    pthread_join( example_threads[current_eti_index], NULL );
    thread_running[ current_eti_index ] = false;

    double join_wait = wait_timer.stop_elapsed();

    timer epoch_timer(true);
    
    // Setup the state for this iteration: the permutations and 
    // point at this iterations pre-sorted examples.
    int *row_perm = row_perms[current_eti_index];
    int *col_perm = col_perms[current_eti_index];
    example_vec** example_vectors = all_example_vectors[current_eti_index];

    DEBUG_ONLY(timer vs_timer(true););
    // *****************************************
    // Now shift around the vectors into their slices.
    for(int k = 0; k < nRows; k++) {
      int row_slice = 0, row_offset = 0;      
      find_slice(L_slice, row_perm[k], row_slice, row_offset);
      if(bTraceNorm) 
	   { store_vec_degree(L_slice, row_slice, row_offset, L[k], L_degree[k]); } 
      else { store_vec(L_slice, row_slice, row_offset, L[k]); }
    }
    // Now shift around the vectors into slices.
    for(int l = 0; l < nCols; l++) {
      int col_slice = 0, col_offset = 0;
      find_slice(R_slice, col_perm[l], col_slice, col_offset);
      if(bTraceNorm) 
	   { store_vec_degree(R_slice, col_slice, col_offset, R[l], R_degree[l]); } 
      else { store_vec(R_slice, col_slice, col_offset, R[l]); }
    }
    DEBUG_ONLY(cout << "\t Vectors Stored in " << vs_timer << endl; );

    // *** BEGIN ITERATION
    DEBUG_ONLY(timer update_timer(true); );
    // We use a cyclic ordering here.
    for(int cycle = 0; cycle < nCol_Slices; cycle++) {      
      pthread_t threads[nRow_Slices];
      DEBUG_ONLY(timer cycle_timer(true););
      for(int k = 0; k < nRow_Slices; k++) { 
	int l = (k + cycle) % nCol_Slices;
	DEBUG_ONLY(timer part_timer(true); );
	// Build the gradient info.
	gradient_thread_info *wti = 
	  bTraceNorm ?
	   new gradient_thread_info(k, L_slice.vecs[k], L_slice.degree_info[k], R_slice.vecs[l], R_slice.degree_info[l], example_vectors[k][l], moviemean, cur_step_size, B, mu, true)
	  :new gradient_thread_info(k, L_slice.vecs[k], NULL, R_slice.vecs[l], NULL, example_vectors[k][l], moviemean, cur_step_size, B, mu, false)
	  ;

	int ret = pthread_create( &threads[k], NULL, gradient_thread_main, (void*) wti);
	if(ret != 0  ) { 
	  cout << "Error in pthread_create: " << ret << endl;
	  exit(-1); 
	}
      }
      for(int i = 0; i < nRow_Slices; i++) {
	pthread_join(threads[i], NULL);
      }
      // Advance to the next eti.      

      DEBUG_ONLY(cout << "\t\t Cycle " << cycle << " took " << cycle_timer << endl;);
    }

    
    DEBUG_ONLY(cout << "\t Computing Update took " << update_timer << endl;);
    // Here, we need to store back
    DEBUG_ONLY(timer store_back(true););
    for(int k = 0; k < nRows; k++) {
      int row_slice = 0, row_offset = 0;
      find_slice(L_slice, row_perm[k], row_slice, row_offset);
      L[k] = get_vec(L_slice, row_slice, row_offset);
    }
    for(int l = 0; l < nCols; l++) {
      int col_slice = 0, col_offset = 0;
      find_slice(R_slice, col_perm[l], col_slice, col_offset);
      R[l] = get_vec(R_slice, col_slice, col_offset);      
    }
    DEBUG_ONLY(cout << "\t Store back took " << store_back  << endl;);    
    // If we need to, we kick off the shuffler here.
    // the current_eti is used up, 
    if(nEpochs - nShufflers > epoch) {
      DEBUG_ONLY(cout << "Kicking off shuffler " << current_eti_index << endl;);
      pthread_create( &example_threads[current_eti_index], NULL, 
		      example_handler_thread_main, 
		      (void*) etis[current_eti_index] );
      thread_running[ current_eti_index ] = true;
    }
    // bookkeeping on end of each epoch
    if ( divergent_series_rule ) {
      nSteps        += nExamples;
      cur_step_size = min(initial_step_size, (nExamples*initial_step_size)/(divergent_series_rule + (double) nSteps));
    } else {
      cur_step_size *= step_diminish;
    }

    // Advance the eti index
    current_eti_index = advance_eti(current_eti_index, number_of_etis);
    
    // End Epoch
    double epoch_duration = epoch_timer.stop_elapsed();
    DEBUG_ONLY(cout << "[TRAIN] epoch " << epoch + 1 << " took " << epoch_timer << " running time: " << whole_jellyfish << endl; );

    // Now compute the RMSEs
    double train_rmse = 0.0, test_rmse = 0.0;
    train_rmse = (nRow_Slices > 1) ? train_tester.test_multithread(m, nRow_Slices) : train_tester.test(m);
    DEBUG_ONLY(cout << "[TRAIN] " << endl; );
    if(t) test_rmse = (nRow_Slices > 1) ? t->test_multithread(m,nRow_Slices) : t->test(m);
    
    if(tp.bPrint) {
      cout << "[Epoch Stats] epoch: " << epoch + 1 << " epoch_time: " << epoch_duration << " join_wait: " << join_wait << " total_time: " << whole_jellyfish << " train_rmse: " << train_rmse << " test_rmse: " << test_rmse << endl;      
    }
  }
  // Thread stops
  for(int i = 0; i < number_of_etis; i++) {
    if( thread_running[i] ) {
      DEBUG_ONLY( cout << " ...  waiting on thread " << i << endl; ) 
      pthread_join( example_threads[i], NULL ); 
      thread_running[i] = false;
    }
  }
  // Cleanup code.
  // This deletes all the scratch for this function.
  for(int i = 0; i < number_of_etis; i++) {
    // quiet down the threads.
    
    delete [] row_perms[i]; delete [] col_perms[i];
    delete [] example_perms[i];

    for(int j = 0; j < nRow_Slices; j++) {delete [] all_example_vectors[i][j];}
    delete [] all_example_vectors[i];
    delete etis[i];
  }
  
  DEBUG_ONLY(cout << "[superfast] finished in " << whole_jellyfish << endl;);
  return m;
}

