/* Authors: Christopher Re and Ben Recht
Copyright 2010-2011 Christopher Re and Ben Recht

This code is released under the GPLv3 License.
http://www.gnu.org/licenses/gpl.html

The files in this directory and its subdirectories are part of Jellyfish.

Jellyifsh is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Jellyfish is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Jellyfish.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "test_handler.h"
#include "pthread.h"
#include "timer.h"
#include <cstdlib>
#include <cmath>

test_handler::
test_handler(const char *szTestFile) {
  int nRows, nCols, nExamples;
  nTests = 0;
    DEBUG_ONLY(timer load_timer; load_timer.start(); );
    DEBUG_ONLY(std::cout << "[TEST] Loading the test set" << std::endl);
    _test = load_examples(szTestFile, nRows, nCols, nExamples);
    _should_free = true;
    nTestExamples = nExamples;
    DEBUG_ONLY(load_timer.stop(); );
    DEBUG_ONLY(std::cout << "[TEST] Finished Loading " << nTestExamples << " examples in " << load_timer.elapsed() << "s" << std::endl);
  }

test_handler::
test_handler(example *test, int nExamples) : nTests(0) {
    _test = test;
    nTestExamples = nExamples;
    _should_free = false;
}


struct 
rmse_helper {
  int start, end;
  model &m;
  test_handler *t;
  double mse;
  rmse_helper(int _start, int _end, 
	      model &_m, test_handler *_t) : m(_m) {
    start = _start;
    end   = _end  ;
    mse   = 0.0   ;
    t     = _t;
  }
};

void*
rmse_run_thread(void *p) {
  struct rmse_helper *h = (struct rmse_helper*) p;
  double error = 0.0;
  test_handler *t = h->t;
  for(int i = h->start ; i < h->end; i++) {
    double v = predict(h->m, t->_test[i].row, t->_test[i].col) - t->_test[i].rating;
    error += v*v;
  }
  h->mse = error;
  return NULL;
}



double 
test_handler::
test(model &m) {
  DEBUG_ONLY(std::cout << "[Test] Starting...[" << nTests++  << "] of " << nTestExamples << std::endl;);  
  double error = 0.0;
  for(int i = 0 ; i < nTestExamples; i++) {
    double v = predict(m, _test[i].row, _test[i].col) - _test[i].rating;
    error += v*v;
  }
  double rmse = sqrt(error)/( sqrt((double) nTestExamples)) ;
  DEBUG_ONLY(std::cout << "[Test] RMSE: " << rmse << std::endl;);
  return rmse;
}

double 
test_handler::
test_multithread(model &m, int nThreads) {
  DEBUG_ONLY(std::cout << "[Test] Starting...[" << nTests++  << " threads=" << nThreads << "]" << std::endl;);  
  double error = 0.0;
  struct rmse_helper* rmse_helpers[nThreads];
  pthread_t helper_threads[nThreads];

  int bucket_size = nTestExamples / nThreads;
  for(int i = 0; i < nThreads; i++) {
    int start = bucket_size * i;
    int end   = std::min(bucket_size * (i+1), nTestExamples);
    rmse_helpers[i] = new rmse_helper(start, end, m, this);
    pthread_create( &(helper_threads[i]), NULL, 
		    rmse_run_thread, rmse_helpers[i]);
  }
  


  for(int i = 0 ; i < nThreads; i++) {     
    pthread_join( helper_threads[i], NULL);
    error += rmse_helpers[i]->mse;
    delete rmse_helpers[i];
  }
  double rmse = sqrt(error)/( sqrt((double) nTestExamples)) ;
  DEBUG_ONLY(std::cout << "[Test] RMSE: " << rmse << std::endl;);
  return rmse;
}
